//
//  LoginView.swift
//  EhPanda
//
//  Created by 荒木辰造 on R 3/08/12.
//

import SwiftUI

struct LoginView: View, StoreAccessor {
    @EnvironmentObject var store: Store
    @Environment(\.dismiss) var dismissAction

    @FocusState private var focusedState: FocusedField?
    @State var shouldRestoreFocus = false
    @State var isLoggingIn = false
    @State var username = ""
    @State var password = ""

    private var isLoginButtonDisabled: Bool {
        username.isEmpty || password.isEmpty
    }
    private var loginButtonColor: Color {
        isLoggingIn ? .clear : isLoginButtonDisabled ? .primary.opacity(0.25) : .primary.opacity(0.75)
    }

    // MARK: LoginView
    var body: some View {
        GeometryReader { proxy in
            ZStack {
                Group {
                    WaveForm(color: Color(.systemGray2).opacity(0.2), amplify: 100, isReversed: true)
                    WaveForm(color: Color(.systemGray).opacity(0.2), amplify: 120, isReversed: false)
                }
                .offset(y: proxy.size.height * 0.3).drawingGroup()
                VStack(spacing: 15) {
                    Group {
                        LoginTextField(
                            focusedState: $focusedState, text: $username, description: "Username", isPassword: false
                        )
                        LoginTextField(
                            focusedState: $focusedState, text: $password, description: "Password", isPassword: true
                        )
                    }
                    .padding(.horizontal, proxy.size.width * 0.2)
                    Button(action: login) {
                        Image(systemName: "chevron.forward.circle.fill")
                    }
                    .overlay { ProgressView().tint(nil).opacity(isLoggingIn ? 1 : 0) }
                    .imageScale(.large).font(.largeTitle).foregroundColor(loginButtonColor)
                    .disabled(isLoginButtonDisabled).padding(.top, 30)
                }
            }
        }
        .toolbar(content: toolbar)
        .onReceive(UIApplication.willResignActiveNotification.publisher) { _ in
            guard focusedState != nil else { return }
            shouldRestoreFocus = true
        }
        .onReceive(UIApplication.didBecomeActiveNotification.publisher) { _ in
            guard shouldRestoreFocus else { return }
            focusedState = .username
            shouldRestoreFocus = false
        }
        .onSubmit {
            switch focusedState {
            case .username:
                focusedState = .password
            case .password:
                focusedState = nil
                login()
            default:
                break
            }
        }
        .navigationTitle("Login").ignoresSafeArea()
    }
    // MARK: Toolbar
    private func toolbar() -> some ToolbarContent {
        ToolbarItem(placement: .navigationBarTrailing) {
            Button {
                store.dispatch(.setSettingViewSheetState(.webviewLogin))
            } label: {
                Image(systemName: "globe")
            }
            .disabled(setting.bypassesSNIFiltering)
        }
    }

    // MARK: Networking
    private func login() {
        guard !isLoginButtonDisabled || isLoggingIn else { return }
        withAnimation { isLoggingIn = true }
        HapticUtil.generateFeedback(style: .soft)

        let token = SubscriptionToken()
        LoginRequest(username: username, password: password)
            .publisher.receive(on: DispatchQueue.main)
            .sink { _ in
                DispatchQueue.main.asyncAfter(deadline: .now() + 3) {
                    withAnimation { isLoggingIn = false }
                    guard AuthorizationUtil.didLogin else {
                        HapticUtil.generateNotificationFeedback(style: .error)
                        return
                    }
                    HapticUtil.generateNotificationFeedback(style: .success)
                    dismissAction.callAsFunction()
                    store.dispatch(.doFinishLoginTasks)
                }
                token.unseal()
            } receiveValue: { value in
                guard setting.bypassesSNIFiltering, let (_, resp) = value as? (Data, HTTPURLResponse) else { return }

                CookiesUtil.setIgneous(for: resp)
                store.dispatch(.fetchIgneous)
            }
            .seal(in: token)
    }
}

// MARK: LoginTextField
private struct LoginTextField: View {
    @Environment(\.colorScheme) private var colorScheme
    private let focusedState: FocusState<FocusedField?>.Binding
    @Binding private var text: String
    private let description: String
    private let isPassword: Bool

    private var backgroundColor: Color {
        colorScheme == .light ? Color(.systemGray6) : Color(.systemGray5)
    }

    init(
        focusedState: FocusState<FocusedField?>.Binding,
        text: Binding<String>, description: String, isPassword: Bool
    ) {
        self.focusedState = focusedState
        _text = text
        self.description = description
        self.isPassword = isPassword
    }

    var body: some View {
        VStack(alignment: .leading, spacing: 8) {
            Text(description.localized).font(.caption).foregroundStyle(.secondary)
            Group {
                if isPassword {
                    SecureField("", text: $text)
                } else {
                    TextField("", text: $text)
                }
            }
            .focused(focusedState.projectedValue, equals: isPassword ? .password : .username)
            .textContentType(isPassword ? .password : .username).submitLabel(isPassword ? .done : .next)
            .textInputAutocapitalization(.none).disableAutocorrection(true).keyboardType(.asciiCapable)
            .padding(10).background(backgroundColor.opacity(0.75).cornerRadius(8))
        }
    }
}

struct LoginView_Previews: PreviewProvider {
    static var previews: some View {
        NavigationView {
            LoginView()
        }
    }
}

// MARK: Definition
private enum FocusedField {
    case username
    case password
}
